/****************************************************************************
 * Grafik-LCD Treiber mit VRAM.
 *
 * (C) Andreas Kaiser 2008.
 *
 * Die Software darf frei kopiert und verndert werden, solange sie
 * nicht ohne meine ausdrckliche Erlaubnis fr kommerzielle Zwecke
 * eingesetzt wird.
 ****************************************************************************/

#include "config.h"
#include "vram.h"
#include "lcd.h"
#include "gbase.h"

bool
g_dot(g_x x, g_y y, g_op op)
{
    if (x >= G_COLS || y >= G_ROWS)
	return false;

    uint8_t mask = g_bitmask[x % LCD_BITS]; x /= LCD_BITS;
    g_write(y, x, op, mask, 0);

    return true;
}

//-----------------------------------------------------------------------------

bool
g_hline(g_x x0, g_x x1, g_y y, g_op op)
{
    if (x0 > x1)
	swap(x0, x1);
    if (x1 >= G_COLS || y >= G_ROWS)
	return false;

    ++x1;
    uint8_t b0 = x0 % LCD_BITS; x0 /= LCD_BITS;
    uint8_t b1 = x1 % LCD_BITS; x1 /= LCD_BITS;

    if (x0 == x1) {
	g_write(y, x0, op, g_setmask[b0] & ~g_setmask[b1], 0);
    } else {
	g_write(y, x0, op, g_setmask[b0], 0);
	g_write(y, x1, op, ~g_setmask[b1], 0);
	while (++x0 < x1) {
	    g_write(y, x0, op, LCD_BITS==4 ? 0xF0 : 0xFF, 0);
	}
    }

    return true;
}

bool
g_vline(g_x x, g_y y0, g_y y1, g_op op)
{
    if (y0 > y1)
	swap(y0, y1);
    if (x >= G_COLS || y1 >= G_COLS)
	return false;

    uint8_t mask = g_bitmask[x % LCD_BITS];
    for (; y0 <= y1; ++y0) {
	g_write(y0, x / LCD_BITS, op, mask, 0);
    }

    return true;
}

bool
g_line(g_x x0, g_y y0, g_x x1, g_y y1, g_op op)
{
    if (y0 == y1)
	return g_hline(x0, x1, y0, op);
    if (x0 == x1)
	return g_vline(x0, y0, y1, op);

    if (x0 >= G_COLS || x1 >= G_COLS || y0 >= G_ROWS || y1 >= G_ROWS)
	return false;

    // Aus Wikipedia: Bresenham Algorithmus

    int16_t x, y, t, dx, dy, incx, incy, pdx, pdy, ddx, ddy, es, el, err;
 
    /* Entfernung in beiden Dimensionen berechnen */
    dx = x1 - x0;
    dy = y1 - y0;
 
    /* Vorzeichen des Inkrements bestimmen */
    if (dx < 0) { incx = -1; dx = -dx; } else if (dx) { incx = +1; } else incx = 0;
    if (dy < 0) { incy = -1; dy = -dy; } else if (dy) { incy = +1; } else incy = 0;

    /* feststellen, welche Entfernung grer ist */
    if (dx > dy) {
	/* x ist schnelle Richtung */
	pdx = incx; pdy = 0;    /* pd. ist Parallelschritt */
	ddx = incx; ddy = incy; /* dd. ist Diagonalschritt */
	es  = dy;   el  = dx;   /* Fehlerschritte schnell, langsam */
    } else {
	/* y ist schnelle Richtung */
	pdx = 0;    pdy = incy; /* pd. ist Parallelschritt */
	ddx = incx; ddy = incy; /* dd. ist Diagonalschritt */
	es  = dx;   el  = dy;   /* Fehlerschritte schnell, langsam */
    }
 
    /* Initialisierungen vor Schleifenbeginn */
    x = x0;
    y = y0;
    err = el / 2;
    g_dot(x, y, op);
 
    /* Pixel berechnen */
    for(t = 0; t < el; ++t) { /* t zaehlt die Pixel, el ist auch Anzahl */
	/* Aktualisierung Fehlerterm */
	err -= es; 
	if (err < 0) {
	    /* Fehlerterm wieder positiv (>=0) machen */
	    err += el;
	    /* Schritt in langsame Richtung, Diagonalschritt */
	    x += ddx;
	    y += ddy;
	} else {
	    /* Schritt in schnelle Richtung, Parallelschritt */
	    x += pdx;
	    y += pdy;
	}
	g_dot(x, y, op);
    }

    return true;
}

//-----------------------------------------------------------------------------

bool
g_fill(g_x x0, g_y y0, g_x x1, g_y y1, g_op op)
{
    if (y0 > y1)
	swap(y0, y1);
    if (x0 >= G_COLS || x1 >= G_COLS || y1 >= G_ROWS)
	return false;

    for (g_y y = y0; y <= y1; ++y) {
	g_hline(x0, x1, y, op);
    }
    return true;
}

//-----------------------------------------------------------------------------

bool
g_data(g_x x, g_y y, g_op op, uint8_t data)
{
    if (x >= G_COLS || y >= G_ROWS)
	return false;

    if (op == G_Replace && (x % 8) == 0) {
	x /= 8;
	if (LCD_BITS == 4) {
	    g_write(y, x+0, G_Replace, 0xF0, data);
	    g_write(y, x+1, G_Replace, 0xF0, data << 4);
	} else {
	    g_write(y, x, G_Replace, 0xFF, data);
	}
    } else {
	for (uint8_t i = 0; i < 8; ++i) {
	    g_dot(x, y, (data & 0x80) ? G_Set : G_Clear);
	    data <<= 1;
	    ++x;
	}
    }

    return true;
}
