/****************************************************************************
 * Grafik-LCD Treiber mit VRAM.
 *
 * (C) Andreas Kaiser 2008.
 *
 * Die Software darf frei kopiert und verndert werden, solange sie
 * nicht ohne meine ausdrckliche Erlaubnis fr kommerzielle Zwecke
 * eingesetzt wird.
 ****************************************************************************/

#include "config.h"
#include "vram.h"
#include "lcd.h"

vram_row_t	lcd_home;	// top row, 0..LCD_ROWS-1

/*-----------------------------------------------------------------------------
 * Timer 0 - Shift clock
 *
 * Clock = CLK/1.
 * CTC mode.
 * OC0(A) toggle.
 * OC0(A) clock = CLK/(2*n).
 */

#if LCD_USE_TIMER

static void
sclk_timer_init(uint8_t n)
{
    setBits(DDR(LCD_SCLK_PORT), LCD_SCLK_PIN);
#ifdef TCCR0A
    TCCR0A = 0<<COM0A1|1<<COM0A0 | 1<<WGM01|0<<WGM00;
    TCCR0B = 0<<WGM02 | 1;
    OCR0A = n;
#else
    TCCR0 = 0<<COM01|1<<COM00 | 1<<WGM01|0<<WGM00 | 1;
    OCR0  = n;
#endif
}

#endif //LCD_USE_TIMER

/*-----------------------------------------------------------------------------
 * Timer 1 - Shift enable
 *
 * Clock = T1 rising edge.
 * Fast PWM mode 0..ICR1.
 * OC1A set on match.
 * Start under software control.
 */

#if LCD_USE_TIMER

static void
sclk_counter_init(uint16_t n)
{
    ICR1   = 0xFFFF;
    OCR1A  = n-1;
    TCNT1  = 0;
    TCCR1A = 1<<COM1A1|1<<COM1A0 | 1<<WGM11|0<<WGM10;
    TCCR1B = 1<<WGM13|1<<WGM12 | 7;
}

static void
sclk_counter_start(void)
{
    TCNT1  = 0xFFFF; // timer overflow resets OC1A
}

#endif //LCD_USE_TIMER

/*-----------------------------------------------------------------------------
 * Timer 2 - Line pulse
 *
 * Mega644 timer 2:
 *	Clock = CLK/8.
 *	Fast PWM mode 0..OC2A.
 *	OC2B set on match.
 * Mega32 timer 2:
 *	Clock = CLK/8.
 *	CTC mode.
 */

static void
lp_timer_init(uint16_t n)
{
#if defined(OCR2B)
    OCR2A  = n-1;
    OCR2B  = n-2; // minimum width (OCR2A=OCR2B: no state change)
    TCCR2A = 1<<COM2B1|1<<COM2B0 | 1<<WGM21|1<<WGM20;
    TCCR2B = 1<<WGM22 | 2;
    TIMSK2 = 1<<OCIE2B;
#else
    OCR2   = n-1;
    TCCR2  = 1<<WGM21|0<<WGM20 | 2;
    TIMSK |= 1<<OCIE2;
#endif
}

#if defined(OCR2B)

#define LP_CLOCK (F_CPU/8)
ISR(TIMER2_COMPB_vect)
{

  #ifdef LCD_M_CYCLE
    static uint8_t current_m_line;
    uint8_t m_line = current_m_line;
    if (m_line == LCD_M_CYCLE) {
	m_line = 0;
	PIN(LCD_M_PORT) = LCD_M_PIN; // toggle
    }
    current_m_line = m_line + 1;
  #endif

#else

#define LP_CLOCK (F_CPU/8)
ISR(TIMER2_COMP_vect)
{
    setBits(PORT(LCD_LP_PORT), LCD_LP_PIN);
    clrBits(PORT(LCD_LP_PORT), LCD_LP_PIN);

  #ifdef LCD_M_CYCLE
    static uint8_t current_m_line;
    uint8_t m_line = current_m_line;
    if (m_line == LCD_M_CYCLE) {
	m_line = 0;
	if (tstBits(PORT(LCD_M_PORT), LCD_M_PIN))
	    clrBits(PORT(LCD_M_PORT), LCD_M_PIN);
	else
	    setBits(PORT(LCD_M_PORT), LCD_M_PIN);
    }
    current_m_line = m_line + 1;
  #endif

#endif

    static uint8_t current_line;
    uint8_t line = current_line;

    static uint8_t current_row;
    vram_row_t row = current_row;

    clrBits(PORT(LCD_FLM_PORT), LCD_FLM_PIN);
    if (line == (LCD_DOUBLE ? LCD_ROWS/2 : LCD_ROWS)) {
	setBits(PORT(LCD_FLM_PORT), LCD_FLM_PIN);
	line = 0;
	row = lcd_home;
    }

    vram_transfer(row, 0);
#if LCD_USE_TIMER
    sclk_counter_start();
#else
  #ifdef LCD_SEN_PIN
    clrBits(PORT(LCD_SEN_PORT), LCD_SEN_PIN);
  #endif
    uint8_t p0 = PORT(LCD_SCLK_PORT);
    uint8_t p1 = p0 ^ LCD_SCLK_PIN;
    uint8_t i = LCD_COLS/LCD_BITS/8;
    do {
    	PORT(LCD_SCLK_PORT) = p1;
	PORT(LCD_SCLK_PORT) = p0;

	PORT(LCD_SCLK_PORT) = p1;
	PORT(LCD_SCLK_PORT) = p0;

	PORT(LCD_SCLK_PORT) = p1;
	PORT(LCD_SCLK_PORT) = p0;

	PORT(LCD_SCLK_PORT) = p1;
	PORT(LCD_SCLK_PORT) = p0;

	PORT(LCD_SCLK_PORT) = p1;
	PORT(LCD_SCLK_PORT) = p0;

	PORT(LCD_SCLK_PORT) = p1;
	PORT(LCD_SCLK_PORT) = p0;

	PORT(LCD_SCLK_PORT) = p1;
	PORT(LCD_SCLK_PORT) = p0;

	PORT(LCD_SCLK_PORT) = p1;
	PORT(LCD_SCLK_PORT) = p0;
    } while (--i);
#endif

    current_row = row + 1;

    vram_refresh();

    current_line = line + 1;

}

//-----------------------------------------------------------------------------

void
lcd_set_home(lcd_row_t row)
{
    cli();
    lcd_home = row;
    sei();
}

//-----------------------------------------------------------------------------

void
lcd_enable(void)
{
    setBits(PORT(LCD_VEN_PORT), LCD_VEN_PIN);
}

//-----------------------------------------------------------------------------

void
lcd_disable(void)
{
    clrBits(PORT(LCD_VEN_PORT), LCD_VEN_PIN);
}

//-----------------------------------------------------------------------------

void
lcd_init(void)
{
    setBits(DDR(LCD_VEN_PORT), LCD_VEN_PIN);	// VEN  is output, initially low
#ifdef LCD_SEN_PIN
    setBits(PORT(LCD_SEN_PORT), LCD_SEN_PIN);	// SEN  is output, initially high
    setBits(DDR(LCD_SEN_PORT), LCD_SEN_PIN);	
#endif
#if !LCD_USE_TIMER && defined(LCD_SEN_PIN)
    setBits(PORT(LCD_SCLK_PORT),LCD_SCLK_PIN);	// SCLK is output, initially high (inverted)
    setBits(DDR(LCD_SCLK_PORT),LCD_SCLK_PIN);
#elif !LCD_USE_TIMER
    clrBits(PORT(LCD_SCLK_PORT),LCD_SCLK_PIN);	// SCLK is output, initially low
    setBits(DDR(LCD_SCLK_PORT),LCD_SCLK_PIN);
#endif
    setBits(DDR(LCD_FLM_PORT), LCD_FLM_PIN);	// FLM  is output  initially low
    setBits(DDR(LCD_LP_PORT),  LCD_LP_PIN);	// LP   is output, initially low
#ifdef LCD_M_CYCLE
    setBits(DDR(LCD_M_PORT),   LCD_M_PIN);	// M    is output, initially low
#endif

#if LCD_USE_TIMER
    sclk_timer_init(1);				// Clock generator, 0=CLK/2,1=CLK/4
    sclk_counter_init(LCD_COLS/LCD_BITS);	// Clock sclk_counter, clocks per line
#endif
    lp_timer_init(LP_CLOCK / (LCD_ROWS * FRAME_RATE)); // LP timer
}

