/****************************************************************************
 * Grafik-LCD Treiber mit VRAM.
 *
 * (C) Andreas Kaiser 2008.
 *
 * Die Software darf frei kopiert und verndert werden, solange sie
 * nicht ohne meine ausdrckliche Erlaubnis fr kommerzielle Zwecke
 * eingesetzt wird.
 ****************************************************************************/

#include "config.h"
#include "usart.h"

#define BAUD UART_BAUD_RATE
#include <util/setbaud.h>

#define Handshake	0				// uses pin change interrupt
#define XonXoff		1

#define TxBuffer	50
#define RxBuffer	250
typedef uint8_t		index_t;			// large enough for Rx and Tx

volatile static index_t	tx_reader, tx_writer, tx_used;	// Tx buffer
volatile static index_t	tx_buffer[TxBuffer];
#if Handshake
volatile static uint8_t	tx_stopped;			// by handshake
#endif
#if XonXoff
volatile static uint8_t	tx_insert, tx_xoff;
#endif

volatile static index_t	rx_reader, rx_writer, rx_used;	// Rx buffer
volatile static index_t	rx_buffer[RxBuffer];
volatile static uint8_t	rx_error;

/*****************************************************************************/

#if Handshake
  #define RTS_init()	DDRC |= 1<<PC2; PORTC &= ~(1<<PC2)
  #define RTS_low()	PORTC &= ~(1<<PC2)
  #define RTS_high()	PORTC |= (1<<PC2)

  #define RxHighWater	(RxBuffer - 10)
  #define RxLowWater	(RxBuffer/2)

  #define CTS_start()	PCMSK1 = 1<<PCINT11; PCICR = 1<<PCIE1
  #define CTS_stop()	PCMSK1 = 0
  #define CTS_high()	(PINC & (1<<PC3))
  #define CTS_interrupt	PCINT1_vect
#endif

#if XonXoff
  #define RxHighWater	((RxBuffer * 3) / 4)
  #define RxLowWater	(RxBuffer / 2)

  #define XON	0x11
  #define XOFF	0x13
#endif

/*****************************************************************************/

void
usartInit(void)
{
    // Rx,Tx enabled, 8N1, interrupts disabled
    UCSR0A = (USE_2X ? 1<<U2X0 : 0) | 0<<MPCM0;
    UCSR0B = 0<<RXCIE0 | 0<<TXCIE0 | 0<<UDRIE0 | 1<<RXEN0 | 1<<TXEN0 | 0<<UCSZ02;
    UCSR0C = 0<<UMSEL01|0<<UMSEL00 | 0<<UPM01|0<<UPM00 | 0<<USBS0 | 1<<UCSZ01|1<<UCSZ00;
    UBRR0  = UBRR_VALUE;
    UCSR0B |= 1<<RXCIE0;	// enable Rx interrupt
#if Handshake
    RTS_init();
    CTS_start();
#endif
}

void
usartStop(void)
{
#if Handshake
    CTS_stop();
#endif
}

void
usartStart(void)
{
    cli();
    rx_reader = rx_writer = rx_used = 0;
    rx_error = 0;
#if Handshake
    tx_stopped = 0;
    RTS_low();
#endif
    sei();
#if XonXoff
    usartTransmit(XON);
    tx_insert = 0;
    tx_xoff = 0;
#endif
}

unsigned
usartError(void)
{
    return rx_error;
}

/*****************************************************************************/

unsigned
usartRxAvail(void)
{
    cli();
    uint16_t n = rx_used;
    sei();
    return n;
}

unsigned
usartReceive(void)
{
    for (;;) {
	cli();
	if (rx_used != 0)
	    break;
	sei();
    }

    uint8_t data = rx_buffer[rx_reader];
    if (++rx_reader == RxBuffer)
	rx_reader = 0;
    --rx_used;

#if XonXoff
    if (tx_xoff && rx_used <= RxLowWater) {
	tx_xoff = 0;
	sei();
	usartTransmit(XON);
    }
#endif

    sei();
    return data;
}

/*****************************************************************************/

unsigned
usartTxAvail(void)
{
    uint8_t sreg = SREG; cli();
    uint16_t n = TxBuffer - tx_used;
    SREG = sreg;
    return n;
}

void
usartTransmit(uint8_t data)
{
    uint8_t sreg = SREG;
    for (;;) {
	cli();
	if (tx_used != TxBuffer)
	    break;
	SREG = sreg;
	if (!(SREG & 0x80))	// do not block when called with interrupts disabled
	    return;
    }

    tx_buffer[tx_writer] = data;
    if (++tx_writer == TxBuffer)
	tx_writer = 0;
    ++tx_used;

    // enable Tx interrupt - high port, need ints disabled
#if Handshake
    if (!tx_stopped)
	UCSR0B |= 1<<UDRIE0;
#else
    UCSR0B |= 1<<UDRIE0;
#endif
    SREG = sreg;
}

/*****************************************************************************/

/*
 * receive complete interrupt
 */
ISR(USART0_RX_vect)
{
    if (rx_used == RxBuffer) {
	(void)UDR0;
	rx_error = 1;
    } else {
	if (UCSR0A & (1<<FE0|1<<DOR0|1<<UPE0))
	    rx_error = 1;

	rx_buffer[rx_writer] = UDR0;
	if (++rx_writer == RxBuffer)
	    rx_writer = 0;
	++rx_used;

#if Handshake
	if (rx_used >= RxHighWater)
	    RTS_high();
	else if (rx_used <= RxLowWater)
	    RTS_low();
#endif
#if XonXoff
	if (rx_used >= RxHighWater) {
	    if (UCSR0B & (1<<UDRIE0))
		tx_insert = XOFF;
	    else
		UDR0 = XOFF;
	    tx_xoff = 1;
	}
#endif
    }
}

/*****************************************************************************/

/*
 * transmit data register empty interrupt
 */
ISR(USART0_UDRE_vect)
{
    if (tx_insert) {
	UDR0 = tx_insert;
	tx_insert = 0;
    } else if (tx_used == 0) {
	UCSR0B &= ~(1<<UDRIE0); // disable UDRE interrupt, do not reset int flag
    } else {
	UDR0 = tx_buffer[tx_reader];
	--tx_used;
	if (++tx_reader == TxBuffer)
	    tx_reader = 0;
    }
}


/*****************************************************************************/
#if BufferedTx && Handshake

/*
 * RTS pin change interrupt
 */
ISR(CTS_interrupt)
{
    if (CTS_high()) {
	tx_stopped = 1;
	UCSR0B &= ~(1<<UDRIE0);			// disable UDRE interrupt
    } else {
	tx_stopped = 0;
	if (tx_used) UCSR0B |= (1<<UDRIE0);	// enable UDRE interrupt
    }
}

#endif
