/****************************************************************************
 * Grafik-LCD Treiber mit VRAM.
 *
 * (C) Andreas Kaiser 2008.
 *
 * Die Software darf frei kopiert und verndert werden, solange sie
 * nicht ohne meine ausdrckliche Erlaubnis fr kommerzielle Zwecke
 * eingesetzt wird.
 ****************************************************************************/

#ifndef __VRAM_H
#define __VRAM_H

#include "config.h"

//-----------------------------------------------------------------------------
// Assume RAS=1,CAS=1,WE=1,OE=1,DSF=0,A8= outside

#define vram_start()				\
    uint8_t __attribute__((unused)) ctrl = VRAM_CTRL_IDLE

#define vram_activate(mask)			\
    ctrl &= ~(mask);				\
    PORT(VRAM_CTRL_PORT) = ctrl

#define vram_deactivate(mask)			\
    ctrl |= (mask);				\
    PORT(VRAM_CTRL_PORT) = ctrl

#define vram_stop()				\
    PORT(VRAM_CTRL_PORT) = VRAM_CTRL_IDLE

#define vram_set_addr(addr)			\
    PORT(VRAM_ADDR_PORT) = addr;

#define vram_data_out()				\
    DDR(VRAM_DATA_PORT) = (VRAM_BITS == 8) ? 0xFF : 0xF0

#define vram_set_data(data)			\
    PORT(VRAM_DATA_PORT) = data

#define vram_data_in()				\
    DDR(VRAM_DATA_PORT) = 0x00

#define vram_get_data()				\
    PIN(VRAM_DATA_PORT)

//-----------------------------------------------------------------------------
// Interrupt functions - not locked

static void __inline __attribute__((always_inline))
vram_transfer(vram_row_t row, vram_col_t col)
{
    vram_start();				//	RAS	CAS	DT/OE	WB/WE
    vram_set_addr(row);				//	1	1	1	1
    vram_activate(VRAM_DTOE | VRAM_RAS);	//	0	1	0	1
    vram_set_addr(col);
    vram_activate(VRAM_CAS);			//	0	0	1	1
    vram_stop();				//	1	1	1	1
}

static void __inline __attribute__((always_inline))
vram_refresh1(void)
{						//	RAS	CAS	DT/OE	WB/WE
    vram_start();		    		//	1	1	1	1
    vram_activate(VRAM_CAS);			//	1	0	1	1
    vram_activate(VRAM_RAS);			//	0	0	1	1
    asm volatile ("nop"); asm volatile ("nop");
    vram_stop();				//	1	1	1	1
}

static void __inline __attribute__((always_inline))
vram_refresh(void)
{
#if VRAM_REFRESH				//	RAS	CAS	DT/OE	WB/WE
    vram_start();		    		//	1	1	1	1
    vram_activate(VRAM_CAS);			//	1	0	1	1
#if VRAM_REFRESH >= 4
    vram_activate(VRAM_RAS);			//	0	0	1	1
    asm volatile ("nop"); asm volatile ("nop");
    vram_deactivate(VRAM_RAS);			//	1	0	1	1
    asm volatile ("nop"); asm volatile ("nop");
#endif
#if VRAM_REFRESH >= 3
    vram_activate(VRAM_RAS);			//	0	0	1	1
    asm volatile ("nop"); asm volatile ("nop");
    vram_deactivate(VRAM_RAS);			//	1	0	1	1
    asm volatile ("nop"); asm volatile ("nop");
#endif
#if VRAM_REFRESH >= 2
    vram_activate(VRAM_RAS);			//	0	0	1	1
    asm volatile ("nop"); asm volatile ("nop");
    vram_deactivate(VRAM_RAS);			//	1	0	1	1
    asm volatile ("nop"); asm volatile ("nop");
#endif
#if VRAM_REFRESH >= 1
    vram_activate(VRAM_RAS);			//	0	0	1	1
    asm volatile ("nop"); asm volatile ("nop");
#endif
    vram_stop();				//	1	1	1	1
#endif
}

//-----------------------------------------------------------------------------
// Application functions - locked

static uint8_t __inline __attribute__((always_inline))
vram_read_i(vram_row_t row, vram_col_t col)
{
    cli();
    vram_start();				//	RAS	CAS	DT/OE	WB/WE
    vram_set_addr(row);				// 	1	1	1	1
    vram_activate(VRAM_RAS);			//	0	1	1	1
    vram_set_addr(col);
    vram_activate(VRAM_CAS | VRAM_DTOE);	//	0	0	0	1
    asm volatile ("nop"); asm volatile ("nop");
    uint8_t data = vram_get_data();
    vram_stop();				// 	1	1	1	1
    sei();
    return data;
}

static void __inline __attribute__((always_inline))
vram_write_i(vram_row_t row, vram_col_t col, uint8_t mask, uint8_t data)
{
    cli();
    vram_start();
    vram_data_out();
    vram_set_addr(row);				//	RAS	CAS	DT/OE	WB/WE
    vram_set_data(mask);			//	1	1	1	1
    vram_activate(VRAM_RAS | VRAM_WBWE);	//	0	1	1	0
    vram_set_addr(col);
    vram_set_data(data);
    vram_activate(VRAM_CAS);			//	0	0	1	0
    vram_stop();				//	1	1	1	1
    vram_data_in();
    sei();
}

//-----------------------------------------------------------------------------

void	vram_init (void);
uint8_t	vram_read (vram_row_t row, vram_col_t col);
void	vram_write (vram_row_t row, vram_col_t col, uint8_t mask, uint8_t data);

#endif //__VRAM_H
